(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Multiset2
imports
  "~~/src/HOL/Library/Multiset"
  Relations 
begin

abbreviation(input) mset_sum where "mset_sum \<equiv> Multiset.fold (op +) (0 :: 'a :: comm_monoid_add)"

lemma comp_fun_commute_plus[simp]: "comp_fun_commute (op + :: (('a :: comm_monoid_add) \<Rightarrow> 'a \<Rightarrow> 'a))"
  by (default, auto simp: ac_simps)

lemma [simp]: "mset_sum (M + N) = mset_sum M + mset_sum N"
  unfolding comp_fun_commute.fold_mset_union [OF comp_fun_commute_plus]
  unfolding comp_fun_commute.fold_mset_fun_left_comm [OF comp_fun_commute_plus]
  by simp

lemma [simp]: "mset_sum {# x #} = x"
proof -
  have "mset_sum ({#x#}) = mset_sum ({#} + {#x#})" by simp
  also have "... = x" 
    unfolding comp_fun_commute.fold_mset_insert[OF comp_fun_commute_plus] by simp
  finally show ?thesis .
qed

lemma union_commutes:
  "M + {#x#} + N = M + N + {#x#}"
  "M + multiset_of xs + N = M + N + multiset_of xs"
  by (auto simp: ac_simps)

lemma in_multiset_of: assumes "x \<in># multiset_of xs"
  shows "\<exists> bef aft. xs = bef @ x # aft"
proof -
  from assms[unfolded in_multiset_in_set set_conv_nth]
  obtain i where i: "i < length xs" and x: "xs ! i = x" by auto
  from id_take_nth_drop[OF i, unfolded x]
  show ?thesis by auto
qed

lemma in_multiset_of_idx :
  assumes "a \<in># multiset_of as"
  shows "\<exists>i. i < length as \<and> a = as ! i"
proof-
 from assms and in_multiset_of[of as a] obtain bef aft where
   as_part: "as = bef @ a # aft" by auto
 with nth_append_length have a_eq: "a = as ! (length bef)" by auto
 from as_part and length_append have i_bd: "(length bef) < length as" by auto
 from a_eq and i_bd show ?thesis by blast
qed

lemma in_set_idx:
  assumes "a \<in> set as"
  shows "\<exists>i. i < length as \<and> a = as ! i"
using assms and in_multiset_in_set[of as a] and in_multiset_of_idx[of as a]
by simp

lemma mem_multiset_diff_compat:
  assumes "x \<in># A"
  and "x \<noteq> y"
  shows "x \<in># (A - {#y#})"
using assms by (induct A) auto

lemma in_multiset_of_eq: assumes "M + {#x#} = multiset_of ys"
  shows "\<exists> bef aft. ys = bef @ x # aft \<and> M = multiset_of (bef @ aft)"
proof -
  from assms[symmetric] have "x \<in># multiset_of ys" by auto
  from in_multiset_of[OF this] obtain bef aft where ys: "ys = bef @ x # aft"
    by auto
  have "M + {#x#} = multiset_of (bef @ aft) + {#x#}"
    unfolding assms ys multiset_eq_iff by auto
  hence "M = multiset_of (bef @ aft)" by auto
  from ys this show ?thesis by auto
qed

lemma multiset_of_concat: assumes "multiset_of xs = multiset_of ys"
  shows "multiset_of (concat xs) = multiset_of (concat ys)"
  using assms
proof (induct xs arbitrary: ys)
  case Nil thus ?case by auto
next
  case (Cons x xs)
  let ?m = "multiset_of"
  from in_multiset_of_eq[OF Cons(2)[simplified]]
  obtain bef aft where ys: "ys = bef @ x # aft" and xs: "?m xs = ?m (bef @ aft)"
    by auto
  from Cons(1)[OF xs] show ?case unfolding ys by (auto simp: multiset_eq_iff)
qed

lemma in_map_multiset_of[intro]:
  "a \<in># A \<Longrightarrow> f a \<in># image_mset f A"
  unfolding mem_set_of_iff[symmetric]
  unfolding set_of_image_mset by auto

definition Union_mset :: "'a multiset multiset \<Rightarrow> 'a multiset"
  where [code del]: "Union_mset \<equiv> Multiset.fold op + {#}"

lemma Union_mset_union[simp]: "Union_mset (M + N) = Union_mset M + Union_mset N"
  unfolding Union_mset_def by simp 

lemma Union_mset_empty[simp]: "Union_mset {#} = {#}" 
  unfolding Union_mset_def by simp

lemma Union_mset_singleton[simp]: "Union_mset {#M#} = M" 
  unfolding Union_mset_def by simp

lemma image_mset_Union_mset: "image_mset f (Union_mset Ms) = Union_mset (image_mset (image_mset f) Ms)"
  by (induct Ms, auto)

lemma Union_mset_mem_subset: assumes "M \<in># ms" shows "M \<le> Union_mset ms"
  using assms
proof (induct ms)
  case empty thus ?case by simp
next
  case (add ms N)
  show ?case
  proof (cases "M = N")
    case True
    thus ?thesis by auto
  next
    case False
    with add(2) have "M \<in># ms" by auto
    from add(1)[OF this] have "M \<le> Union_mset ms" .
    also have "... \<le> Union_mset ms + Union_mset {#N#}"
      by auto
    finally show ?thesis by simp
  qed
qed

lemma multiset_of_replicate_mono: assumes "m1 \<le> m2"
  shows "Union_mset (multiset_of (replicate n m1))
    \<le> Union_mset (multiset_of (replicate n m2))"
proof (induct n)
  case (Suc n)
  have "Union_mset (multiset_of (replicate (Suc n) m1)) = 
    Union_mset (multiset_of (replicate n m1)) + m1" by simp
  also have "\<dots> \<le> Union_mset (multiset_of (replicate n m1)) + m2" using `m1 \<le> m2` by auto
  also have "\<dots> \<le> Union_mset (multiset_of (replicate n m2)) + m2" using Suc by auto
  finally show ?case by auto
qed simp

lemma multised_add_singleton[simp]: "{# a #} + M = N + {# a #} \<longleftrightarrow> M = N"
  "M + {# a #} = {# a #} + N \<longleftrightarrow> M = N"
  by (auto simp: ac_simps)

lemma multiset_image_set_mono: assumes AB: "A \<le> B" shows "image_mset f A \<le> image_mset f B"
proof -
  from AB[unfolded mset_le_exists_conv] obtain C where B: "B = A + C" ..
  thus ?thesis by auto
qed

lemma mem_Union_mset[simp]: "x \<in># Union_mset ms \<longleftrightarrow> x \<in> \<Union> (set_of ` (set_of ms))"
  by (induct ms, auto)
  
lemma diff_union_swap_multiset :
  "a \<in># A \<Longrightarrow> A - {#a#} + B = A + B - {#a#}"
  by (intro multiset_eqI) auto

text{* In a multiset, there exists a local maxima. *}

lemma local_maxima_multiset:
  assumes "locally_trans s A A A" (is "?lt A")
    and "locally_irrefl s A" (is "?li A")
    and A_ne: "A \<noteq> {#}"
  shows "\<exists>a. a \<in># A \<and> (\<forall>b. b \<in># A \<longrightarrow> (b,a) \<notin> s)"
using assms
proof (induct A)
  case empty thus ?case by auto
next
  case (add A x)
  from lt_trans_l[OF `?lt (A + {# x #})`] have lt_A: "?lt A" .
  from li_trans_l[OF `?li (A + {# x #})`] have li_A: "?li A" .
  note IH = add(1)[OF lt_A li_A]
  show ?case
  proof (cases "A = {#}")
    case True
    with `?li (A + {# x #})`[unfolded locally_irrefl_def, rule_format] show ?thesis
    by simp
  next
    case False
    from IH[OF this]
    obtain a1 where
      a1_in_A: "a1 \<in># A"
      and a1_is_lm: "\<forall>b. b \<in># A \<longrightarrow> (b,a1) \<notin> s" by force
    show ?thesis
    proof (cases "(x,a1) \<in> s")
      case True note sxa1 = this
      have GL: "x \<in># A + {# x #}" by auto
      from add(3)[unfolded locally_irrefl_def]
      have GR1: "(x,x) \<notin> s" by simp
      {
        fix b assume b_in_A: "b \<in># A"
        have "(b,x) \<notin> s"
        proof (cases "(b,x) \<in> s")
          case True
          from b_in_A have b_in: "b \<in># A + {# x #}" by simp
          from a1_in_A have a1_in: "a1 \<in># A + {# x #}" by simp
          from b_in and a1_in and GL and True and sxa1 and
            add(3)[unfolded locally_irrefl_def] and
            add(2)[unfolded locally_trans_def]
          have "(b,a1) \<in> s"  by metis
          with a1_is_lm and b_in_A show ?thesis by auto
        qed 
      }
      hence GR2: "\<forall>b. b \<in># A \<longrightarrow> (b,x) \<notin> s" by simp
      from this and GL and GR1 show ?thesis by auto
    next
      case False
      from this and a1_is_lm
      have GR: "\<forall>b. b \<in># A + {# x #} \<longrightarrow> (b,a1) \<notin> s" by simp
      from a1_in_A have GL: "a1 \<in># A + {# x #}" by simp
      from GL and GR show ?thesis by metis
    qed
  qed
qed

lemma mult1_ab_a: "({#a#}, {#a,b#}) \<in> mult1 r"
  unfolding mult1_def
proof(rule, unfold split, intro exI conjI)
  show "{#a, b#} = {#a#} + {#b #}"
    by (simp add: ac_simps)
next
  show "{#a#} = {# a #} + {#}" by simp
qed simp

lemma mult1_ab_b: "({#b#}, {#a,b#}) \<in> mult1 r"
proof -
  have id: "{#a,b#} = {#b,a#}" by (simp add: ac_simps)
  show ?thesis unfolding id by (rule mult1_ab_a)
qed

declare mult1_union[intro!]

lemma mult1_mono_right[intro!]: assumes "(M,N) \<in> mult1 r"
  shows "((M + T, N + T) \<in> mult1 r)"
  using mult1_union[OF assms, of T] by (simp add: ac_simps)

lemma mult1_singleton[simp]: 
  shows "((M, {# t #}) \<in> mult1 r) = (\<forall> a. a \<in># M \<longrightarrow> (a,t) \<in> r)" (is "?l = ?r")
proof
  assume r: ?r
  show ?l
    unfolding mult1_def
    by (rule, unfold split, rule exI[of _ t], rule exI[of _ "{#}"], 
      rule exI[of _ M], insert r, auto)
next
  assume l: ?l
  from this[unfolded mult1_def]
  obtain a M0 K where id: "{#t#} = M0 + {#a#}"
    "M = M0 + K" and rel: "\<forall> b. b \<in># K \<longrightarrow> (b,a) \<in> r" by blast
  from id(1)[unfolded single_is_union] have "M0 = {#}" by simp
  with id have id: "t = a" "M = K" by auto
  show ?r unfolding id by (rule rel)
qed

inductive_set bounded_mult :: "nat \<Rightarrow> 'a rel \<Rightarrow> 'a multiset rel" for b and r where
  bounded_mult: "J \<noteq> {#} \<Longrightarrow> size K \<le> b \<Longrightarrow>
    (\<And> k. k \<in> set_of K \<Longrightarrow> \<exists>j \<in> set_of J. (k, j) \<in> r) \<Longrightarrow> (I + K, I + J) \<in> bounded_mult b r"

lemma bounded_mult_into_mult: assumes trans: "trans r"
  shows mem: "(M,N) \<in> bounded_mult b r \<Longrightarrow> (M,N) \<in> mult r"
  by (induct rule: bounded_mult.induct, intro one_step_implies_mult[OF trans], auto)

definition bmult_less where "bmult_less b \<equiv> bounded_mult b {(x,y :: nat). x < y}"
definition bound_mset :: "nat \<Rightarrow> nat multiset \<Rightarrow> nat" where
  "bound_mset b \<equiv> \<lambda> M. mset_sum (image_mset (\<lambda> x. (Suc b) ^ x) M)"

declare bound_mset_def[simp]

lemma bound_mset_linear_in_size: 
  assumes ij: "\<And> j. j \<in> set_of J \<Longrightarrow> j < i"
  shows "bound_mset b J = 0 \<or> (\<exists> x. x < i \<and> bound_mset b J \<le> size J * (Suc b^x))"
  using ij
proof (induct J rule: multiset_induct)
  case (add J x) 
  show ?case
  proof (cases "size J = 0")
    case True
    thus ?thesis using add by (intro disjI2 exI[of _ x], auto)
  next
    case False
    let ?bb = "Suc b"
    let ?bound_mset = "bound_mset b"
    from add(1)[OF add(2)] False obtain y where yi: "y < i" and bound: "?bound_mset J \<le> size J * ?bb ^ y" by (cases J, auto)
    let ?x = "max x y"
    from yi add(2)[of x] have xi: "?x < i" by auto
    have "y \<le> ?x" by simp
    from power_increasing[OF this, of ?bb] 
    have "size J * ?bb ^ y \<le> size J * ?bb ^ ?x" by simp
    with bound have *: "?bound_mset J \<le> size J * ?bb ^ ?x" by arith
    have "x \<le> ?x" by simp
    from power_increasing[OF this, of ?bb] *
    have "?bound_mset (J + {# x #}) \<le> size (J + {# x #}) * ?bb ^ ?x" by simp
    with xi show ?thesis by blast
  qed
qed simp

lemma bound_mset_bmult_less: "(M,N) \<in> bmult_less b ^^ k \<Longrightarrow> k \<le> bound_mset b N"
proof (induct k arbitrary: M N)
  case (Suc k)
  let ?bound_mset = "bound_mset b"
  from Suc(2) obtain K where mk: "(M,K) \<in> bmult_less b ^^ k" and kn: "(K,N) \<in> bmult_less b" by auto
  from Suc(1)[OF mk] have k: "k \<le> ?bound_mset K" .
  show ?case using kn unfolding bmult_less_def
  proof (cases)
    case (bounded_mult I J L)
    from bounded_mult(3,5) have "\<exists> i. i \<in> set_of I \<and> (\<forall> j \<in> set_of J. i > j)"
    proof (induct J rule: multiset_induct)
      case empty
      thus ?case by (cases I, auto)
    next
      case (add J j)
      from add(3)[of j] obtain i1 where i1: "i1 \<in> set_of I" and i1j: "i1 > j" by auto
      from add(1)[OF add(2) add(3)] obtain i2 where i2: "i2 \<in> set_of I" and i2J: "\<And> j. j \<in> set_of J \<Longrightarrow> i2 > j" by force
      show ?case
      proof (cases "i1 < i2")
        case True
        thus ?thesis
          by (intro exI[of _ i2], insert i2 i2J i1j, auto)
      next
        case False
        {
          fix j
          assume "j \<in> set_of J"
          from i2J[OF this] False have "i1 > j" by simp
        }
        thus ?thesis
          by (intro exI[of _ i1], insert i1 i1j, auto)
      qed
    qed
    then obtain i where i: "i \<in> set_of I" and ij: "\<And> j. j \<in> set_of J \<Longrightarrow> i > j" by blast
    def bb \<equiv> "Suc b"
    have bb1: "1 \<le> bb" unfolding bb_def by simp
    from i have I: "I = I - {# i #} + {# i #}" by auto
    from arg_cong[OF this, of ?bound_mset] 
    have le: "bb^i \<le> ?bound_mset I" unfolding bb_def by simp
    from bound_mset_linear_in_size[OF ij, of J b]
    have choice: "?bound_mset J = 0 \<or> (\<exists> x. x < i \<and> ?bound_mset J \<le> size J * (bb^x))" unfolding bb_def .
    have "?bound_mset J + 1 \<le> bb^i"
    proof (cases "?bound_mset J = 0")
      case True
      thus ?thesis by (simp add: bb_def)
    next
      case False
      with choice obtain x where xi: "x < i" and bound: "?bound_mset J \<le> size J * bb^x" by auto
      have "size J * bb ^ x \<le> b * bb ^ x" using bound bounded_mult(4) by simp
      with bound have bound: "?bound_mset J \<le> b * bb^x" by arith
      also have "b * bb^x < (b+1) * bb^x" using bb1 by simp
      also have "\<dots> = bb^(x + 1)" unfolding bb_def by simp
      also have "\<dots> \<le> bb^i" by (rule power_increasing[OF _ bb1], insert xi, auto)
      finally show ?thesis by simp
    qed
    with le      
    have "?bound_mset J + 1 \<le> ?bound_mset I" by arith
    thus ?thesis using k unfolding bounded_mult(1-2) 
      by simp
  qed
qed simp

declare bound_mset_def[simp del]

end
