(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)

header {* Systems of Representatives of Equivalence Relations *}

theory Representative_System
imports Main
begin

definition repsys :: "'a set \<Rightarrow> 'a rel \<Rightarrow> 'a set"
where
  "repsys A R = {(SOME x. x \<in> X) | X. X \<in> A // R}"

text {*Equivalence classes of an equivalence relation are nonempty.*}
lemma class_nonempty:
  assumes "equiv A R" and "X \<in> A // R"
  shows "\<exists>x. x \<in> X"
  using assms by (metis equiv_class_self quotientE)

lemma repsys_subset:
  assumes "equiv A R"
  shows "repsys A R \<subseteq> A"
  using class_nonempty [OF assms]
  by (auto simp: repsys_def) (metis UnionI Union_quotient assms someI_ex)

lemma mem_repsys_class:
  assumes "equiv A R" and "x \<in> repsys A R"
  shows "\<exists>X \<in> A // R. x \<in> X"
  using class_nonempty [OF assms(1)] and assms(2)
  by (auto simp: repsys_def) (metis someI_ex)

lemma repsys_unique_modulo:
  assumes "equiv A R"
    and "x \<in> repsys A R" and "y \<in> repsys A R" and "x \<noteq> y"
  shows "(x, y) \<notin> R"
  using assms by (auto simp: repsys_def) (metis (lifting) class_nonempty quotient_eq_iff someI_ex)

lemma repsys_representative:
  assumes "equiv A R"
    and "x \<in> A"
  shows "\<exists>y \<in> repsys A R. (x, y) \<in> R"
proof -
  from Union_quotient [OF assms(1)] and `x \<in> A` obtain X
    where "X \<in> A // R" and "x \<in> X" by auto
  then have "(x, (SOME x. x \<in> X)) \<in> R"
    using quotient_eq_iff [OF assms(1)] by (metis someI_ex)
  moreover have "(SOME x. x \<in> X) \<in> repsys A R"
    using `X \<in> A // R` by (auto simp: repsys_def)
  ultimately show ?thesis by blast
qed

end

