(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory TermFun
imports TermFun_Setup
begin

text {* Below, we provide a rewriting view for some predefined concepts *}

text {* Register common types: *}

setup {*
  fold Term_Fun.register_datatypes
   (map single ["HOL.bool", "Nat.nat", "List.list", "Option.option"])
*}

text {*
  Assume globally that everything lives in the class.
  Constraints can be removed later...
*}

default_sort trs_rep

text {* Define TRSs for some standard functions. *}

define_trs plus_nat.simps
define_trs minus_nat.simps
define_trs take.simps
define_trs drop.simps
lemma list_len_easy: "length (x # xs) = Suc (length xs)" by simp
define_trs list.size(3) list_len_easy

fun le :: "nat \<Rightarrow> nat \<Rightarrow> bool" where
  "le 0 n = True"
| "le (Suc n) 0 = False"
| "le (Suc n) (Suc m) = le n m"

define_trs le.simps

lemma le_subst[trs_subst]: "x \<le> y = (le x y)"
  by (induct x y rule: le.induct) auto

fun eq where
  "eq 0 0 = True"
| "eq (Suc x) 0 = False"
| "eq 0 (Suc x) = False"
| "eq (Suc x) (Suc y) = eq x y"

define_trs eq.simps

lemma eq_subst[trs_subst]: "(x = y) = eq x y"
  by (induct x y rule: eq.induct) auto


fun less where
  "less 0 0 = False"
| "less (Suc x) 0 = False"
| "less 0 (Suc x) = True"
| "less (Suc x) (Suc y) = less x y"

define_trs less.simps

lemma less_subst[trs_subst]: "(x < y) = less x y"
  by (induct x y rule: less.induct) auto

end
